/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

#define USE_XIM

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <gtk/gtkentry.h>
#include <gtk/gtksignal.h>
#include <gdk/gdkprivate.h>
#include <X11/Xlib.h>
#include "gtkpandaentry.h"

static void gtk_panda_entry_class_init    (GtkPandaEntryClass *klass);
static void gtk_panda_entry_init          (GtkPandaEntry     *entry);

static void gtk_panda_entry_realize       (GtkWidget         *widget);
static void gtk_panda_entry_unrealize     (GtkWidget         *widget);
static void gtk_panda_entry_hide          (GtkWidget         *widget);
static gint gtk_panda_entry_button_press  (GtkWidget         *widget,
					   GdkEventButton    *event);
static gint gtk_panda_entry_focus_in      (GtkWidget         *widget,
					   GdkEventFocus     *event);
static gint gtk_panda_entry_focus_out     (GtkWidget         *widget,
					   GdkEventFocus     *event);
static void gtk_panda_entry_insert_text   (GtkEditable       *editable,
					   const gchar       *new_text,
					   gint               new_text_length,
					   gint              *position);

static GtkEntryClass *parent_class = NULL;

GtkType
gtk_panda_entry_get_type (void)
{
  static GtkType type = 0;

  if (!type)
    {
      static const GtkTypeInfo info =
      {
	"GtkPandaEntry",
	sizeof (GtkPandaEntry),
	sizeof (GtkPandaEntryClass),
	(GtkClassInitFunc) gtk_panda_entry_class_init,
	(GtkObjectInitFunc) gtk_panda_entry_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };

      type = gtk_type_unique (GTK_TYPE_ENTRY, &info);
    }

  return type;
}

static void
gtk_panda_entry_class_init (GtkPandaEntryClass *class)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;
  GtkEditableClass *editable_class;
  GtkEntryClass *entry_class;

  object_class = (GtkObjectClass*) class;
  widget_class = (GtkWidgetClass*) class;
  editable_class = (GtkEditableClass*) class;
  entry_class = (GtkEntryClass*) class;
  parent_class = gtk_type_class (GTK_TYPE_ENTRY);

  widget_class->realize = gtk_panda_entry_realize;
  widget_class->unrealize = gtk_panda_entry_unrealize;
  widget_class->hide = gtk_panda_entry_hide;
  widget_class->button_press_event = gtk_panda_entry_button_press;
  widget_class->focus_in_event = gtk_panda_entry_focus_in;
  widget_class->focus_out_event = gtk_panda_entry_focus_out;

  editable_class->insert_text = gtk_panda_entry_insert_text;
}

static void
gtk_panda_entry_init (GtkPandaEntry *entry)
{
  entry->input_mode = GTK_PANDA_ENTRY_XIM_MODE;
  entry->xim_enabled = FALSE;
}

GtkWidget*
gtk_panda_entry_new (void)
{
  return GTK_WIDGET (gtk_type_new (GTK_PANDA_TYPE_ENTRY));
}

void
gtk_panda_entry_set_input_mode (GtkPandaEntry *entry,
				enum gtk_panda_entry_input_mode mode)
{
  entry->input_mode = mode;
}

void
gtk_panda_entry_set_xim_enabled (GtkPandaEntry *entry, gboolean flag)
{
  entry->xim_enabled = flag;
}

static GdkWChar
gtk_entry_get_invisible_char (GtkEntry *entry)
{
  GdkWChar ch = 0;

  if (entry->use_wchar)
    gdk_mbstowcs (&ch, "*", 1);
  else
    ch = '*';

  return ch;
}

/* Recompute the x offsets of all characters in the buffer */
static void
gtk_entry_recompute_offsets (GtkEntry *entry)
{
  gint i;
  gint offset = 0;
  GtkEditable *editable = GTK_EDITABLE (entry);

  for (i=0; i<entry->text_length; i++)
    {
      GdkWChar ch;

      entry->char_offset[i] = offset;

      if (editable->visible)
	ch = entry->text[i];
      else
	ch = gtk_entry_get_invisible_char (entry);

      if (entry->use_wchar)
	offset += gdk_char_width_wc (GTK_WIDGET (entry)->style->font, ch);
      else
	offset += gdk_char_width (GTK_WIDGET (entry)->style->font, ch);
    }
  
  entry->char_offset[i] = offset;
}

static void
gtk_panda_entry_realize (GtkWidget *widget)
{
  GtkEntry *entry;
  GtkEditable *editable;
  GtkRequisition requisition;
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_ENTRY (widget));

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
  entry = GTK_ENTRY (widget);
  editable = GTK_EDITABLE (widget);

  gtk_widget_get_child_requisition (widget, &requisition);
  
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y + (widget->allocation.height -
					 requisition.height) / 2;
  attributes.width = widget->allocation.width;
  attributes.height = requisition.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget);
  attributes.event_mask |= (GDK_EXPOSURE_MASK |
			    GDK_BUTTON_PRESS_MASK |
			    GDK_BUTTON_RELEASE_MASK |
			    GDK_BUTTON1_MOTION_MASK |
			    GDK_BUTTON3_MOTION_MASK |
			    GDK_POINTER_MOTION_HINT_MASK |
			    GDK_ENTER_NOTIFY_MASK |
			    GDK_LEAVE_NOTIFY_MASK |
			    GDK_KEY_PRESS_MASK);
  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

  widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), &attributes, attributes_mask);
  gdk_window_set_user_data (widget->window, entry);

  attributes.x = widget->style->klass->xthickness;
  attributes.y = widget->style->klass->ythickness;
  attributes.width = widget->allocation.width - attributes.x * 2;
  attributes.height = requisition.height - attributes.y * 2;
  attributes.cursor = entry->cursor = gdk_cursor_new (GDK_XTERM);
  attributes_mask |= GDK_WA_CURSOR;

  entry->text_area = gdk_window_new (widget->window, &attributes, attributes_mask);
  gdk_window_set_user_data (entry->text_area, entry);

  widget->style = gtk_style_attach (widget->style, widget->window);

  gdk_window_set_background (widget->window, &widget->style->base[GTK_WIDGET_STATE (widget)]);
  gdk_window_set_background (entry->text_area, &widget->style->base[GTK_WIDGET_STATE (widget)]);

  gdk_window_show (entry->text_area);

  if (editable->selection_start_pos != editable->selection_end_pos)
    gtk_editable_claim_selection (editable, TRUE, GDK_CURRENT_TIME);

  gtk_entry_recompute_offsets (entry);
}

static void
gtk_panda_entry_unrealize (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_ENTRY (widget));

  if (GTK_WIDGET_CLASS (parent_class)->unrealize)
    (* GTK_WIDGET_CLASS (parent_class)->unrealize) (widget);
}

static void
gtk_panda_entry_hide (GtkWidget *widget)
{
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_WIDGET (widget));

  if (GTK_WIDGET_CLASS (parent_class)->hide)
    (* GTK_WIDGET_CLASS (parent_class)->hide) (widget);

#ifdef USE_XIM
  if (GTK_EDITABLE (widget)->ic)
    {
      gdk_ic_destroy (GTK_EDITABLE (widget)->ic);
      GTK_EDITABLE (widget)->ic = NULL;
    }
  if (GTK_EDITABLE (widget)->ic_attr)
    {
      gdk_ic_attr_destroy (GTK_EDITABLE (widget)->ic_attr);
      GTK_EDITABLE (widget)->ic_attr = NULL;
    }
#endif
}

static gint
gtk_panda_entry_button_press (GtkWidget      *widget,
			      GdkEventButton *event)
{
  if (event->button == 4 || event->button == 5)
    return TRUE;

  return GTK_WIDGET_CLASS (parent_class)->button_press_event (widget, event);
}

static gint
gtk_panda_entry_focus_in (GtkWidget     *widget,
			  GdkEventFocus *event)
{
  GtkEntry *entry;
  GtkEditable *editable;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_ENTRY (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  entry = GTK_ENTRY (widget);
  editable = GTK_EDITABLE (widget);

#ifdef USE_XIM
  if ((GTK_PANDA_ENTRY (widget)->input_mode == GTK_PANDA_ENTRY_XIM_MODE
       || GTK_PANDA_ENTRY (widget)->input_mode == GTK_PANDA_ENTRY_KANA_MODE)
      && editable->ic == NULL
      && gdk_im_ready ()
      && (editable->ic_attr = gdk_ic_attr_new ()) != NULL)
    {
      gint width, height;
      GdkEventMask mask;
      GdkColormap *colormap;
      GdkICAttr *attr = editable->ic_attr;
      GdkICAttributesType attrmask = GDK_IC_ALL_REQ;
      GdkIMStyle style;
      GdkIMStyle supported_style = GDK_IM_PREEDIT_NONE |
				   GDK_IM_PREEDIT_NOTHING |
			           GDK_IM_PREEDIT_POSITION |
			           GDK_IM_STATUS_NONE |
				   GDK_IM_STATUS_NOTHING;

      if (widget->style && widget->style->font->type != GDK_FONT_FONTSET)
	supported_style &= ~GDK_IM_PREEDIT_POSITION;

      attr->style = style = gdk_im_decide_style (supported_style);
      attr->client_window = entry->text_area;

      if ((colormap = gtk_widget_get_colormap (widget)) !=
	    gtk_widget_get_default_colormap ())
	{
	  attrmask |= GDK_IC_PREEDIT_COLORMAP;
	  attr->preedit_colormap = colormap;
	}
      attrmask |= GDK_IC_PREEDIT_FOREGROUND;
      attrmask |= GDK_IC_PREEDIT_BACKGROUND;
      attr->preedit_foreground = widget->style->fg[GTK_STATE_NORMAL];
      attr->preedit_background = widget->style->base[GTK_STATE_NORMAL];

      switch (style & GDK_IM_PREEDIT_MASK)
	{
	case GDK_IM_PREEDIT_POSITION:
	  if (widget->style && widget->style->font->type != GDK_FONT_FONTSET)
	    {
	      g_warning ("over-the-spot style requires fontset");
	      break;
	    }

	  gdk_window_get_size (entry->text_area, &width, &height);

	  attrmask |= GDK_IC_PREEDIT_POSITION_REQ;
	  attr->spot_location.x = 0;
	  attr->spot_location.y = height;
	  attr->preedit_area.x = 0;
	  attr->preedit_area.y = 0;
	  attr->preedit_area.width = width;
	  attr->preedit_area.height = height;
	  attr->preedit_fontset = widget->style->font;

	  break;
	}
      editable->ic = gdk_ic_new (attr, attrmask);
     
      if (editable->ic == NULL)
	g_warning ("Can't create input context.");
      else
	{
	  mask = gdk_window_get_events (entry->text_area);
	  mask |= gdk_ic_get_events (editable->ic);
	  gdk_window_set_events (entry->text_area, mask);

	  if (GTK_WIDGET_HAS_FOCUS(widget))
	    gdk_im_begin (editable->ic, entry->text_area);
	}
    }
#endif

  if (GTK_WIDGET_CLASS (parent_class)->focus_in_event)
    (* GTK_WIDGET_CLASS (parent_class)->focus_in_event) (widget, event);

#ifdef USE_XIM
  if (GTK_PANDA_ENTRY (widget)->xim_enabled && gdk_xim_ic && gdk_xim_ic->xic)
    {
      XVaNestedList *preedit_attr =
	XVaCreateNestedList (0, XNPreeditState, XIMPreeditEnable, NULL);
      XSetICValues (gdk_xim_ic->xic, XNPreeditAttributes, preedit_attr, NULL);
      XFree (preedit_attr);
    }
#endif

  return FALSE;
}

static gint
gtk_panda_entry_focus_out (GtkWidget     *widget,
			   GdkEventFocus *event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_ENTRY (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (GTK_WIDGET_CLASS (parent_class)->focus_out_event)
    (* GTK_WIDGET_CLASS (parent_class)->focus_out_event) (widget, event);

  return FALSE;
}

static gchar *
get_symbol (gchar key)
{
  static gchar *symbol_table[] = {
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", 0, 0, 0, "", 0, 0, 0, "", 0, 0, 0, 0, 0, "",
    0, 0, 0, 0, 0, "", 0, 0, 0, 0, 0, "", "", "", "", 0
  };

  if (key >= 0x20)
    return symbol_table[key - 0x20];
  return 0;
}

static gchar *
get_kana (gchar *prefix, gchar key)
{
  static struct {
    gchar *prefix;
    gchar *a, *i, *u, *e, *o;
  } kana_table[] = {
    {"l",  "", "", "", "", ""},
    {"x",  "", "", "", "", ""},
    {"k",  "", "", "", "", ""},
    {"ky", "", "", "", "", ""},
    {"g",  "", "", "", "", ""},
    {"gy", "", "", "", "", ""},
    {"s",  "", "", "", "", ""},
    {"sh", "", "", "", "", ""},
    {"sy", "", "", "", "", ""},
    {"z",  "", "", "", "", ""},
    {"j",  "", "", "", "", ""},
    {"jy", "", "", "", "", ""},
    {"zy", "", "", "", "", ""},
    {"t",  "", "", "", "", ""},
    {"ts", 0, 0, "", 0, 0},
    {"lt", 0, 0, "", 0, 0},
    {"xt", 0, 0, "", 0, 0},
    {"ty", "", "", "", "", ""},
    {"cy", "", "", "", "", ""},
    {"ch", "", "", "", "", ""},
    {"th", "ƥ", "ƥ", "ƥ", "ƥ", "ƥ"},
    {"d",  "", "", "", "", ""},
    {"dy", "¥", "¥", "¥", "¥", "¥"},
    {"dh", "ǥ", "ǥ", "ǥ", "ǥ", "ǥ"},
    {"dw", "ɥ", "ɥ", "ɥ", "ɥ", "ɥ"},
    {"n",  "", "", "", "", ""},
    {"ny", "˥", "˥", "˥", "˥", "˥"},
    {"h",  "", "", "", "", ""},
    {"f",  "ե", "ե", "", "ե", "ե"},
    {"hy", "ҥ", "ҥ", "ҥ", "ҥ", "ҥ"},
    {"fy", "ե", "ե", "ե", "ե", "ե"},
    {"b",  "", "", "", "", ""},
    {"by", "ӥ", "ӥ", "ӥ", "ӥ", "ӥ"},
    {"p",  "", "", "", "", ""},
    {"py", "ԥ", "ԥ", "ԥ", "ԥ", "ԥ"},
    {"m",  "", "", "", "", ""},
    {"my", "ߥ", "ߥ", "ߥ", "ߥ", "ߥ"},
    {"y",  "", "", "", "", ""},
    {"ly", "", 0, "", 0, ""},
    {"xy", "", 0, "", 0, ""},
    {"r",  "", "", "", "", ""},
    {"ry", "", "ꥣ", "", "ꥧ", ""},
    {"w",  "", "", "", "", ""},
    {"wh", "", "", "", "", ""},
    {"lw", "", 0, 0, 0, 0},
    {"xw", "", 0, 0, 0, 0},
    {"v",  "", "", "", "", ""},
    {"q",  "", "", "", "", ""},
    {0, 0, 0, 0, 0, 0}
  };

  int i;
  for (i = 0; kana_table[i].prefix; i++)
    if (strcmp (prefix, kana_table[i].prefix) == 0)
      switch (key)
	{
	case 'a': return kana_table[i].a;
	case 'i': return kana_table[i].i;
	case 'u': return kana_table[i].u;
	case 'e': return kana_table[i].e;
	case 'o': return kana_table[i].o;
	default:  return 0;
	}
  return 0;
}

static void
gtk_panda_entry_insert_text (GtkEditable *editable,
			     const gchar *new_text,
			     gint         new_text_length,
			     gint        *position)
{
  GtkEntry *entry;

  g_return_if_fail (editable != NULL);
  g_return_if_fail (GTK_IS_EDITABLE (editable));

  entry = GTK_ENTRY (editable);

  /* Handle kana input */
  if (GTK_PANDA_ENTRY (entry)->input_mode == GTK_PANDA_ENTRY_KANA_MODE)
    {
      gchar buff[8];
      gint prefix_start, prefix_end;
      gchar *prefix, this, last;

      /* Do nothing if there is no input */
      if (new_text_length < 1)
	return;

      /* Just insert for non-keyboard input */
      if (new_text_length > 1)
	goto insert;

      /* Find the prefix */
      prefix_end = editable->current_pos;
      for (prefix_start = prefix_end; prefix_start > 0; prefix_start--)
	{
	  gchar *s = gtk_editable_get_chars (editable, prefix_start - 1, -1);
	  if (*s < 'a' || 'z' < *s)
	    break;
	}
      prefix = gtk_editable_get_chars (editable, prefix_start, prefix_end);

      this = *new_text;
      last = *prefix;

      if (last == 0)
	{
	  /* single char */
	  gchar *s = get_symbol (new_text[0]);
	  if (s != 0)
	    new_text = s;
	  else
	    goto insert;
	}
      else if (this == 'a' || this == 'i' || this == 'u'
	       || this == 'e' || this == 'o')
	{
	  /*  */
	  gchar *s = get_kana (prefix, new_text[0]);
	  if (s != 0)
	    new_text = s;
	  else
	    goto insert;
	}
      else if (last == 'n' && this != 'y')
	{
	  /* n ->  */
	  if (this == 'n' || this == '\'')
	    strcpy (buff, "");
	  else
	    sprintf (buff, "%c", this);
	  new_text = buff;
	}
      else if (this == last)
	{
	  /* xx ->  */
	  sprintf (buff, "%c", this);
	  new_text = buff;
	}
      else
	goto insert;

      new_text_length = strlen (new_text);
      *position = prefix_start;
      gtk_editable_delete_text (editable, prefix_start, prefix_end);
    }

 insert:
  if (GTK_EDITABLE_CLASS (parent_class)->insert_text)
    (* GTK_EDITABLE_CLASS (parent_class)->insert_text)
      (editable, new_text, new_text_length, position);
}
