/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GTK+ Team and others 1997-1999.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gtkpandaps.h"

#include "zoom-in.xpm"
#include "zoom-out.xpm"

static void gtk_panda_ps_class_init    (GtkPandaPSClass *klass);
static void gtk_panda_ps_init          (GtkPandaPS      *pps);

GtkType
gtk_panda_ps_get_type (void)
{
  static GtkType type = 0;

  if (!type)
    {
      static const GtkTypeInfo info =
      {
	"GtkPandaPS",
	sizeof (GtkPandaPS),
	sizeof (GtkPandaPSClass),
	(GtkClassInitFunc) gtk_panda_ps_class_init,
	(GtkObjectInitFunc) gtk_panda_ps_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };

      type = gtk_type_unique (gtk_vbox_get_type (), &info);
    }

  return type;
}

static void
gtk_panda_ps_class_init (GtkPandaPSClass *class)
{
}

static void
gtk_panda_ps_init (GtkPandaPS *pps)
{
  pps->gs = NULL;
  pps->pan = 0;
  pps->prev_x = pps->prev_y = 0;
}

static void
button_press_callback (GtkWidget *widget, GdkEventButton *event, gpointer data)
{
  GtkPandaPS *pps = data;
  gint button = event->button;
  gint x = 0;
  gint y = 0;

  x += event->x;
  y += event->y;
  switch (button)
    {
    case 1:
      if (!pps->pan)
	{
	  static GdkCursor *pan_cursor = NULL;
	  gint wx = 0, wy = 0;

	  gdk_window_get_pointer (widget->window, &wx, &wy, NULL);

	  pps->pan = TRUE;
	  if (pan_cursor == NULL)
	    pan_cursor = gdk_cursor_new (GDK_FLEUR);

	  gtk_grab_add (widget);
	  gdk_pointer_grab (widget->window, FALSE,
			    GDK_POINTER_MOTION_MASK |
			    GDK_BUTTON_RELEASE_MASK, NULL,
			    pan_cursor, GDK_CURRENT_TIME);
	  pps->prev_x = wx;
	  pps->prev_y = wy;
	  gtk_gs_start_scroll (GTK_GS (pps->gs));
	}
      break;
    case 2:
    case 3:
      break;
      /* Support for wheel mice */
    case 4:
      if (event->state & GDK_CONTROL_MASK)
	gtk_gs_scroll (GTK_GS (pps->gs), -64, 0);
      else
	gtk_gs_scroll (GTK_GS (pps->gs), 0, -64);
      break;
    case 5:
      if (event->state & GDK_CONTROL_MASK)
	gtk_gs_scroll (GTK_GS (pps->gs), 64, 0);
      else
	gtk_gs_scroll (GTK_GS (pps->gs), 0, 64);
      break;
    default:
      break;
    }
}

static void
button_release_callback (GtkWidget *widget, GdkEventButton *event, gpointer data)
{
  GtkPandaPS *pps = data;

  switch (event->button)
    {
    case 1:
      if (pps->pan)
	{
	  pps->pan = FALSE;
	  gdk_pointer_ungrab (GDK_CURRENT_TIME);
	  gtk_grab_remove (widget);
	  gtk_gs_end_scroll (GTK_GS (pps->gs));
	}
      break;
    default:
      break;
    }
}

static void
motion_callback (GtkWidget *widget, GdkEventMotion *event, gpointer data)
{
  GtkPandaPS *pps = data;

  if (pps->pan)
    {
      gtk_gs_scroll (GTK_GS (pps->gs),
		     -event->x + pps->prev_x,
		     -event->y + pps->prev_y);
      pps->prev_x = event->x;
      pps->prev_y = event->y;
    }
}

void
zoom_in (GtkWidget *widget, gpointer data)
{
  GtkPandaPS *pps = GTK_PANDA_PS (data);
  float z = gtk_gs_get_zoom (GTK_GS (pps->gs)) * 1.2;
  gtk_gs_set_zoom (GTK_GS (pps->gs), z);
}

void
zoom_out (GtkWidget *widget, gpointer data)
{
  GtkPandaPS *pps = GTK_PANDA_PS (data);
  float z = gtk_gs_get_zoom (GTK_GS (pps->gs)) / 1.2;
  gtk_gs_set_zoom (GTK_GS (pps->gs), z);
}

static GtkWidget *
append_to_toolbar (GtkWidget *toolbar, gchar **xpm_data, const char *tooltip,
		   GtkSignalFunc callback, gpointer user_data)
{
  GdkColormap *colormap;
  GdkPixmap *pixmap;
  GdkBitmap *mask;
  GtkWidget *icon;

  colormap = gtk_widget_get_colormap (toolbar);
  pixmap = gdk_pixmap_colormap_create_from_xpm_d (NULL, colormap, &mask,
						  NULL, xpm_data);
  icon = gtk_pixmap_new (pixmap, mask);

  return gtk_toolbar_append_element (GTK_TOOLBAR (toolbar),
				     GTK_TOOLBAR_CHILD_BUTTON,
				     NULL, "", tooltip, NULL,
				     icon, callback, user_data);
}

GtkWidget *
gtk_panda_ps_new (void)
{
  GtkWidget *frame1;
  GtkWidget *toolbar, *zoomin, *zoomout;
  GtkPandaPS *pps = GTK_PANDA_PS (gtk_type_new (GTK_PANDA_TYPE_PS));
  GtkObject *hadj = gtk_adjustment_new (0.0, 0.0, 1.0, 1.0, 1.0, 0.5);
  GtkObject *vadj = gtk_adjustment_new (0.0, 0.0, 1.0, 1.0, 1.0, 0.5);

  /* GtkGS */
  pps->gs = gtk_gs_new (GTK_ADJUSTMENT (hadj), GTK_ADJUSTMENT (vadj));
  gtk_widget_show (pps->gs);
  gtk_widget_set_events (pps->gs,
			 GDK_BUTTON_PRESS_MASK
			 | GDK_BUTTON_RELEASE_MASK
			 | GDK_POINTER_MOTION_MASK);
  gtk_signal_connect (GTK_OBJECT (pps->gs), "button_press_event",
		      GTK_SIGNAL_FUNC (button_press_callback), pps);
  gtk_signal_connect (GTK_OBJECT (pps->gs), "button_release_event",
		      GTK_SIGNAL_FUNC (button_release_callback), pps);
  gtk_signal_connect (GTK_OBJECT (pps->gs), "motion_notify_event",
		      GTK_SIGNAL_FUNC (motion_callback), pps);

  /* zoom-in/out buttons */
  toolbar = gtk_toolbar_new (GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_ICONS);
  zoomout = append_to_toolbar (toolbar, zoom_out_xpm, "Zoom out", zoom_out, pps);
  zoomin = append_to_toolbar (toolbar, zoom_in_xpm, "Zoom in", zoom_in, pps);
  gtk_widget_show (toolbar);
  gtk_widget_show (zoomout);
  gtk_widget_show (zoomin);

  /* frame */
  frame1 = gtk_frame_new (NULL);
  gtk_frame_set_shadow_type (GTK_FRAME (frame1), GTK_SHADOW_IN);
  gtk_widget_show (frame1);
  gtk_container_add (GTK_CONTAINER (frame1), pps->gs);

  GTK_BOX (pps)->spacing = 4;
  GTK_BOX (pps)->homogeneous = FALSE;
  gtk_box_pack_start (GTK_BOX (pps), toolbar, FALSE, FALSE, 0);
  gtk_box_pack_start (GTK_BOX (pps), frame1, TRUE, TRUE, 0);

  return GTK_WIDGET (pps);
}

void
gtk_panda_ps_load (GtkPandaPS *pps, const char *filename)
{
  if (gtk_gs_load (GTK_GS (pps->gs), filename))
    {
      float z;
      gtk_gs_set_page_size(GTK_GS (pps->gs), -1, 0);
      z = gtk_gs_zoom_to_fit (GTK_GS (pps->gs), TRUE);
      if (0.002 < z && z < 2000)
	gtk_gs_set_zoom (GTK_GS (pps->gs), z);
    }
}
